/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.noderoute;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.tinygroup.binarytree.AVLTree;
import org.tinygroup.binarytree.impl.AVLTreeImpl;

/**
 * 路由表
 * 
 * @author luoguo
 * 
 */
public final class NodeRouteTable implements NodeRoute {
	public static final boolean DIRECT_SINGLE = false;// 单向
	public static final boolean DIRECT_DOUBLE = true;// 双向
	/**
	 * 当前节点
	 */
	private Node currentNode;
	/**
	 * 根据node名称来确定其Node节点
	 */
	private Map<String, Node> nodeMap = new HashMap<String, Node>();
	/**
	 * 与某一节点直接相连的节点列表
	 */
	private Map<Node, AVLTree<Node>> connectNodeMap = new HashMap<Node, AVLTree<Node>>();
	/**
	 * 分组名对应的子节点
	 */
	private Map<String, List<Node>> groupNodes = new HashMap<String, List<Node>>();
	/**
	 * 路由表 Node为目标节点 AVLTree<Node>为路由表
	 */
	private Map<Node, Map<Node, List<Node>>> routeMap = new HashMap<Node, Map<Node, List<Node>>>();

	public NodeRouteTable() {
	}

	public Map<Node, AVLTree<Node>> getConnectNodeMap() {
		return connectNodeMap;
	}

	public void setConnectNodeMap(Map<Node, AVLTree<Node>> connectNodeMap) {
		this.connectNodeMap = connectNodeMap;
	}

	public NodeRouteTable(Node node) {
		this.currentNode = node;
	}

	public Node getCurrentNode() {
		return currentNode;
	}

	public void setCurrentNode(Node currentNode) {
		this.currentNode = currentNode;
	}

	public Map<String, List<Node>> getGroupNodes() {
		return groupNodes;
	}

	public void setGroupNodes(Map<String, List<Node>> groupNodes) {
		this.groupNodes = groupNodes;
	}

	public Map<String, Node> getNodeMap() {
		return nodeMap;
	}

	public void setNodeMap(Map<String, Node> nodeMap) {
		this.nodeMap = nodeMap;
	}

	public Map<Node, Map<Node, List<Node>>> getRouteMap() {
		return routeMap;
	}

	public void setRouteMap(Map<Node, Map<Node, List<Node>>> routeMap) {
		this.routeMap = routeMap;
	}

	/**
	 * 添加双向节点
	 * 
	 * @param startNode
	 * @param endNode
	 */
	public void addNode(Node startNode, Node endNode) {
		addNode(startNode, endNode, true);
	}

	/**
	 * 添加节点
	 * 
	 * @param startNode
	 * @param endNode
	 * @param isTwoway
	 *            是否双向，true时为双向
	 */
	public void addNode(Node startNode, Node endNode, boolean isTwoway) {
		if (nodeMap.get(startNode.getName()) == null) {
			putToNodeMap(startNode);
		}
		if (nodeMap.get(endNode.getName()) == null) {
			putToNodeMap(endNode);
		}
		putToConnectMap(startNode, endNode, isTwoway);
		putToGroupMap(startNode);
		putToGroupMap(endNode);
		rebuildRouteTable();
	}

	public void addNode(Node node) {
		addNode(currentNode, node);
	}

	private void putToGroupMap(Node node) {
		String groupName = node.getGroupName();
		if (groupName != null && groupName.length() > 0) {
			List<Node> nodes = groupNodes.get(groupName);
			if (nodes == null) {
				nodes = new ArrayList<Node>();
			}
			if (nodes.indexOf(node) < 0) {
				nodes.add(node);
			}
		}
	}

	private void putToConnectMap(Node startNode, Node endNode) {
		AVLTree<Node> nodes = connectNodeMap.get(startNode);
		if (nodes == null) {
			nodes = new AVLTreeImpl<Node>();
			connectNodeMap.put(startNode, nodes);
		}
		nodes.add(endNode);
	}

	private void putToConnectMap(Node startNode, Node endNode, boolean isTwoway) {
		putToConnectMap(startNode, endNode);
		if (isTwoway) {
			putToConnectMap(endNode, startNode);
		}
	}

	private void putToNodeMap(Node node) {
		nodeMap.put(node.getName(), node);
	}

	private void rebuildRouteTable() {
		for (Node startNode : nodeMap.values()) {
			for (Node endNode : nodeMap.values()) {
				if (!startNode.equals(endNode)) {
					rebuildRouteTable(startNode, endNode);
				}
			}
		}
	}

	/**
	 * 重新计算路由表 算法思路： Start->后找一下，如果找到，则结束 End->向前找一下，如果找到，则结束 如果没有的话，则继续
	 * 
	 */

	private void rebuildRouteTable(Node startNode, Node endNode) {
		List<Node> startList = new ArrayList<Node>();
		startList.add(startNode);
		List<Node> endList = new ArrayList<Node>();
		endList.add(endNode);
		rebuildRouteTable(startNode, endNode, startList, endList);
	}

	private void rebuildRouteTable(Node startNode, Node endNode,
			List<Node> startList, List<Node> endList) {
		AVLTree<Node> connectStartNode = connectNodeMap.get(startNode);
		if (connectStartNode.contains(endNode) != null) {
			List<Node> routeList = combineList(startList, endList);
			addRouteTable(startNode, endNode, routeList);
		}
	}

	private void addRouteTable(Node startNode, Node endNode,
			List<Node> routeList) {
		Map<Node, List<Node>> nodeRouteMap = routeMap.get(startNode);
		if (nodeRouteMap == null) {
			nodeRouteMap = new HashMap<Node, List<Node>>();
			routeMap.put(startNode, nodeRouteMap);
		}
		nodeRouteMap.put(endNode, routeList);

	}

	private List<Node> combineList(List<Node> startList, List<Node> endList) {
		List<Node> combineList = new ArrayList<Node>();
		combineList.addAll(startList);
		combineList.addAll(endList);

		return combineList;

	}

	private int dijkstra(Node startNode, Node endNode) {
		Map<Node, Boolean> isLabel = new HashMap<Node, Boolean>();// 是否标号
		Node[] indexs = new Node[nodeMap.size()];// 所有标号的点的下标集合，以标号的先后顺序进行存储，实际上是一个以数组表示的栈
		int i_count = -1;// 栈的顶点
		Map<Node, Integer> distance = new HashMap<Node, Integer>();// v0到各点的最短距离的初始值
		Node index = startNode;// 从初始点开始
		int presentShortest = 0;// 当前临时最短距离

		indexs[++i_count] = index;// 把已经标号的下标存入下标集中
		isLabel.put(index, true);
		Collection<Node> nodes = nodeMap.values();

		while (i_count < nodeMap.size()) {
			// 第一步：标号v0,即w[0][0]找到距离v0最近的点

			int min = Integer.MAX_VALUE;
			for (Node i : nodes) {
				if (isLabel.get(i) != null && !isLabel.get(i)
						&& distance.get(i) != null && i != index) {
					// 如果到这个点有边,并且没有被标号
					if (distance.get(i) < min) {
						min = distance.get(i);
						index = i;// 把下标改为当前下标
					}
				}
			}
			if (index.equals(endNode)) {// 已经找到当前点了，就结束程序
				break;
			}
			isLabel.put(index, true);// 对点进行标号
			indexs[++i_count] = index;// 把已经标号的下标存入下标集中

			Integer dis = distance.get(index);
			if (dis == null) {
				dis = -1;
			}
			if (connectNodeMap.get(indexs[i_count - 1]).contains(index) != null
					|| presentShortest + 1 > dis) {
				// 如果两个点没有直接相连，或者两个点的路径大于最短路径
				presentShortest = dis;
			} else {
				presentShortest += 1;
			}

			// 第二步：将distance中的距离加入vi
			for (Node node : distance.keySet()) {
				// 如果vi到那个点有边，则v0到后面点的距离加
				if (distance.get(node) == -1
						&& connectNodeMap.get(index).contains(node) != null) {// 如果以前不可达，则现在可达了
					distance.put(node, presentShortest + 1);
				} else if (connectNodeMap.get(index).contains(node) != null
						&& presentShortest + 1 < distance.get(node)) {
					// 如果以前可达，但现在的路径比以前更短，则更换成更短的路径
					distance.put(node, presentShortest + 1);
				}
			}
		}
		// 如果全部点都遍历完，则distance中存储的是开始点到各个点的最短路径
		return distance.get(endNode) - distance.get(startNode);
	}
}
